// Java Interface to Illustra

// Copyright (c) 1997. Erwin Kalvelagen, TNO FEL, <kalvelagen@fel.tno.nl>
// Copyright (c) 1996 Bradley McLean / Jeffrey Medeiros
// Modifications Copyright (c) 1996/1997 Martin Rode
// Copyright (c) 1997 Peter T Mount
//
// You may distribute under the terms of the GNU Public License as
// specified in the README file that comes with the JavaPostgres95 Kit

package mijdbc;

import java.sql.*;
import java.math.*;
import java.net.*;
import java.io.*;
//import java.util.zip.*;

/**
 * JDBC Interface to Illustra functions
 */

public class MIConnection implements Connection
{


  // these constants are for use with Illustra
  final static int DBCLONG = 0x352468bd;     // long word ordering info
  final static int DBCSHORT = 0xac01;        // short word ordering info
  final static int DB_PROTOCOL = 6;          // protocol level
  final static String LANGUAGE = "english";  // database system language
  final static String ENCODING = "ASCII";    // character encoding

  // socket options
  final static boolean TCP_NODELAY = false;  // not needed
  //final static int BUFSIZE = 8192; // buffer size for input
  final static int BUFSIZE = 0; // buffer size for input

  // if talking to bridge, turn on compression
  boolean COMPRESS = true; 

  MIDriver parent;  

  MITypeTable typetable; // we maintain a type table

  String Host;    // host address, either hostname or dotted decimal number
  int Port;       // port number; note: default Illustra port is 7531
  String Options;
  String Tty;     // display (not passed on yet; but illustra can handle it)
  String Dbname;  // database to connect to
  String User;    // username
  String Pass;    // password


  // socket connections. Notice that for output we use a buffered
  // stream, so always use flush when appropriate.
  DataInputStream sock_in;
  DataOutputStream sock_out;

  boolean autoCommit = true;
  boolean ReadOnly = false;
  String Catalog = null;
  static final int TRANSACTION_NONE		 = 0;
  static final int TRANSACTION_READ_UNCOMMITTED = 1;
  static final int TRANSACTION_READ_COMMITTED   = 2;
  static final int TRANSACTION_REPEATABLE_READ  = 3;
  static final int TRANSACTION_SERIALIZABLE     = 4;
  static final int TRANSACTION_VERSIONING       = 5;
  int transactionIsolation = TRANSACTION_NONE;
  boolean autoClose = true;
  SQLWarning warning = null;
  


  private void connect() throws SQLException {
    
    // 
    // EK: this routine is complete rewritten as Postgres does not resemble
    // Illustra here.
    //
    
    try {
      
      
      // open socket connection
      Socket sock = new Socket(Host,Port);

      // turn off buffering
      if (TCP_NODELAY)
      	sock.setTcpNoDelay(true);

      // set timeout
      //sock.setSoTimeout(10000);



      // input stream
      InputStream is = sock.getInputStream();
      if (BUFSIZE > 0)
	is = new BufferedInputStream(is, BUFSIZE);
      if (COMPRESS)
	is = new MIInflaterInputStream(is);
      if (MId.dump)
	is = new MIBufferedInputStream(is, MIStatement.MAX_MESSAGE_LEN);
      sock_in = new DataInputStream(is);

      
      // output stream is buffered. Flushing is already called where needed!
      // MIBufferedOutputStream will hex dump outgoing traffic
      if (MId.dump)
  	sock_out = new DataOutputStream(new MIBufferedOutputStream(sock.getOutputStream(),
								   MIStatement.MAX_MESSAGE_LEN));
      else
	sock_out = new DataOutputStream(new BufferedOutputStream(sock.getOutputStream(),
								 MIStatement.MAX_MESSAGE_LEN));
      
      
      // Send packet:
      //     word32 containing long word ordering (DBCLONG = 0x352468bd)
      //     word16 containing short word ordering (DBCSHORT = 0xac01)
      //     char8  class 'C'
      //     char8  subclass 'C'
      sock_out.writeInt(DBCLONG);
      sock_out.writeShort(DBCSHORT);
      sock_out.writeByte('C');
      sock_out.writeByte('C');
      sock_out.flush();
      
      // Send packet:
      //    Header:
      //       word32 containing length payload = 4
      //       word16 containing channel = 0 (not used)
      //       char8 class 'C'
      //       char8 subclass 'P'
      //    Payload:
      //       word32 protocol version, DB_PROTOCOL=6
      
      sock_out.writeInt(4);
      sock_out.writeShort(0);
      sock_out.writeByte('C');
      sock_out.writeByte('P');
      sock_out.writeInt(DB_PROTOCOL);
      sock_out.flush();
      
      // Receive packet
      //    Header:
      //       word32 containing length payload = 4
      //       word16 containing channel = 0 (not used)
      //       char8 class 'C'
      //       char8 subclass 'P'
      //    Payload:
      //       word32 protocol version, DB_PROTOCOL=6 hopefully
      
      
      sock_in.readInt();
      sock_in.readShort();
      char c = (char) sock_in.readByte();
      char d = (char) sock_in.readByte();
      int protocol = sock_in.readInt();
      if (c != 'C' || d != 'P' || protocol != DB_PROTOCOL)
	throw new SQLException("Protocol negotiation failed");
      
      
    } catch ( Exception e ) {
      throw new SQLException("Server Connection Open Failed: "+e);
    }
    
  } // connect
  

   private void login() throws SQLException {

     // Illustra login procedure

     try {
       
       // send packet:
       //    Header:
       //       word32 containing length payload = 4 + length(username) + 
       //                                          4 + length(password) + 
       //                                          4 + length(tty)
       //       word16 containing channel = 0 (not used)
       //       char8 class 'C'
       //       char8 subclass 'A'
       //    Payload:
       //       word32: length of username
       //       chars: username
       //       word32: length of password
       //       chars: password
       //       word32: length of terminal name 
       //       chars: terminal name 

       sock_out.writeInt(12 + User.length() + Pass.length() + Tty.length());
       sock_out.writeShort(0);
       sock_out.writeByte('C');
       sock_out.writeByte('A');
       sock_out.writeInt(User.length());
       sock_out.writeBytes(User);
       sock_out.writeInt(Pass.length());
       sock_out.writeBytes(Pass);
       sock_out.writeInt(Tty.length());
       sock_out.writeBytes(Tty);

       // no need to flush. add one more message:

       // send packet:
       //    Header:
       //       word32 containing length payload = 12 + dbname.length + language.length + encoding.length
       //       word16 containing channel = 0 (not used)
       //       char8 class 'A'
       //       char8 subclass 'D'
       //    Payload:
       //       word32: length of dbname
       //       chars: dbname
       //       word32: length of language
       //       chars: language
       //       word32: length of encoding
       //       chars: encoding

       sock_out.writeInt(12 + Dbname.length() + LANGUAGE.length() + ENCODING.length());
       sock_out.writeShort(0);
       sock_out.writeByte('A');
       sock_out.writeByte('D');
       sock_out.writeInt(Dbname.length());
       sock_out.writeBytes(Dbname);
       sock_out.writeInt(LANGUAGE.length());
       sock_out.writeBytes(LANGUAGE);
       sock_out.writeInt(ENCODING.length());
       sock_out.writeBytes(ENCODING);
       sock_out.flush(); // needed here...


       // Receive packet
       //    Header:
       //       word32 containing length payload 
       //       word16 containing channel = 0 (not used)
       //       char8 class 
       //       char8 subclass
       //    Payload:
       //        length payload bytes....
       //
       // Packet ('Q','R') means Ready for Query
       // Packet ('C','D') means login failed.
       // Packet ('Q','E') indicates failure
       // Everything else is unexpected.....

       int payloadlength = sock_in.readInt();
       sock_in.readShort();
       char c = (char) sock_in.readByte();
       char d = (char) sock_in.readByte();
       while (payloadlength > 0)
	 payloadlength -= sock_in.skipBytes(payloadlength);

       if (c ==  'Q' && d == 'R') {
	 if (MId.debug) 
	   MId.h("Login succeeded");
	 return;  // this means OK
       }

       if (c == 'C' && d == 'D') 
	 throw new SQLException("Login Failed");

       if (c == 'Q' && d == 'E')
	 throw new SQLException("Login Error");

       throw new SQLException("Unknown response");

     } catch ( Exception e ) {
       throw new SQLException("Login Failed: "+e);
     }
     

   } // login


  /**
   * Establish connection with Illustra server (mideamon/miserver)
   */

   MIConnection(MIDriver up, String mihost, String miport, String mioptions,
                	String mitty,String dbname,String miuser,
                        String mipass) throws SQLException {
      parent = up;

      Host = mihost;
      Port = Integer.valueOf(miport).intValue();
      Options = mioptions;
      Tty = mitty;
      Dbname = dbname;
      User = miuser;
      Pass = mipass;

      typetable = new MITypeTable();
      
      if (MId.debug) {
	System.out.println("Host = "+Host);
	System.out.println("Port = "+Port);
	System.out.println("Dbname = "+Dbname);
	System.out.println("User = "+User);
	System.out.println("Tty = "+Tty);
      }

      if (Port == 7532) 
	COMPRESS = true;	
      else 
	COMPRESS = false;

      // connect to database
      connect();

      // for Illustra we need to login....
      login();

   } // MIConnection


   // createStatement returns an SQL Statement object
   public Statement createStatement() throws SQLException {
      return new MIStatement(this);
   }
   // prepareStatement creates a pre-compiled SQL PreparedStatement object.
   public PreparedStatement prepareStatement(String sql) throws SQLException {
      return new MIPreparedStatement(this,sql);
   }

   // prepareCall create a pre-compiled SQL statement that is
   // a call on a stored procedure.
   public CallableStatement prepareCall(String sql) throws SQLException {
      return new MICallableStatement(this,sql);
   }

   // Convert the given generic SQL statement to the drivers native SQL.
   public String nativeSQL(String query) throws SQLException {
System.out.println("nativeSQL called");
      return query;
   }

   // If "autoCommit" is true, then all subsequent SQL statements will
   // be executed and committed as individual transactions.  Otherwise
   // (if "autoCommit" is false) then subsequent SQL statements will
   // all be part of the same transaction , which must be explicitly
   // committed with either a "commit" or "rollback" call.
   // By default new connections are initialized with autoCommit "true".
   public void setAutoCommit(boolean enableAutoCommit) throws SQLException {
      if ( (! autoCommit) && enableAutoCommit )
         commit();
      autoCommit = enableAutoCommit;
      if ( ! autoCommit)
         begin();
   }

   public boolean getAutoCommit() throws SQLException
   {
      return false;
   }

   // You can use commit or rollback to commit or abort a transaction.
   // Note that by default all active Statement or ResultSet objects
   // associated with a Connection will be closed when it is committed
   // or aborted.
   protected void begin() throws SQLException {
//      ((MIStatement)createStatement()).exec("begin");
   }

   public void commit() throws SQLException {
//      ((MIStatement)createStatement()).exec("commit");
   }

   public void rollback() throws SQLException {
//      ((MIStatement)createStatement()).exec("rollback");
   }

   // close frees up various state associated with the connection.
   public void close() throws SQLException {
      if ( ! autoCommit )
         commit();
      // close connection
      

      // send packet:
      //    Header:
      //       word32 containing length payload = 0
      //       word16 containing channel = 0 (not used)
      //       char8 class 'C'
      //       char8 subclass 'D'

      try {
       sock_out.writeInt((int)0);
       sock_out.writeShort((short)0);
       sock_out.writeByte('C');
       sock_out.writeByte('D');
       sock_out.flush(); // needed here...
      } catch ( Exception e ) {
	throw new SQLException("Sending close down packket failed "+e);
      }
      
      parent.closeConnection(this);
   }

   // isClosed returns true if the connection is closed, which can
   // occur either due to an explicit call on "close" or due to
   // some fatal error on the connection.
   public boolean isClosed() throws SQLException {
      return false;
   }

   //======================================================================
   // Advanced features:

   // You can obtain a DatabaseMetaData object to get information
   // about the target database.
   public DatabaseMetaData getMetaData() throws SQLException {
      return new MIDatabaseMetaData(this);
   }

   // You can put a connection in read-only mode as a hint to enable
   // database optimizations.  Note that setReadOnly cannot be called
   // while in the middle of a transaction.
   public void setReadOnly(boolean readOnly) throws SQLException {
      ReadOnly=readOnly;
   }
   public boolean isReadOnly() throws SQLException {
      return ReadOnly;
   }

   // The "catalog" selects a sub-space of the target database.
   public void setCatalog(String catalog) throws SQLException {
      Catalog = catalog;
   }
   public String getCatalog() throws SQLException {
      return Catalog;
   }

   // You can call the following method to try to change the transaction
   // isolation level on a newly opened connection, using one of the
   // TRANSACTION_* values.  Use the DatabaseMetaData class to find what
   // isolation levels are supported by the current database.
   // Note that setTransactionIsolation cannot be called while in the
   // middle of a transaction.
   public void setTransactionIsolation(int level) throws SQLException {
      if ( level < TRANSACTION_NONE || level > TRANSACTION_VERSIONING )
         throw new SQLException();
      transactionIsolation = level;
   }
   public int getTransactionIsolation() throws SQLException {
      return transactionIsolation;
   }

   // disableAutoClose can be used on a newly opened connection to to
   // disable the automatic closing of Statements and ResultSets during
   // calls on Commit and Rollback.  This functionality is only available
   // in some databases, and the exact semantics of the objects state
   // vary between databases.  See the DatabaseMetaData class for
   // methods to learn about the consequences of doing disableAutoClose.
   public void disableAutoClose() throws SQLException {
      autoClose = false;
   }

   public void setAutoClose(boolean autoClose) throws SQLException
   {
   }

   public boolean getAutoClose() throws SQLException
   {
      return false;
   }

   // getWarnings will return any warning information related to
   // the current connection.  Note that SQLWarning may be a chain.
   public SQLWarning getWarnings() throws SQLException {
      return warning;
   }
   public void clearWarnings() throws SQLException {
      warning = null;
   }
}

