// Java Interface to Illustra
// Erwin Kalvelagen, TNO FEL

// Copyright (c) 1996 Bradley McLean / Jeffrey Medeiros
// Modifications Copyright (c) 1996/1997 Martin Rode
// Copyright (c) 1997 Peter T Mount
//
// You may distribute under the terms of the GNU Public License as
// specified in the README file that comes with the JavaPostgres95 Kit

// PM Peter Mount <pmount@maidast.demon.co.uk>
// TG Thorsten Greiner <thorsten.greiner@consol.de>

// Column Names are stored in 'columns'
// Values are stored in 'tuples'

package mijdbc;

import java.sql.*;
import java.math.*;
import java.util.*;
import java.text.*;
import java.io.*;

/**
 * JDBC Interface to Illustra functions
 */

public class MIResultSet implements ResultSet
{
  MIStatement parent;
  Vector columns;
  Vector tuples;
  int currentRow;
  private Hashtable months;
  private boolean wasException;
  private MIResultSetMetaData meta;
  private MITypeTable TypeTable;
  //long tupletime;

  byte[] flags;


  /**
   * Helper function to read a Tuple from the server
   * This functions has to run fast!!!
   */
  private Object[] readTuple(DataInputStream in, int nfields, boolean binary, int packetlen) throws SQLException
  {


    Object[] data = new Object[nfields];

    try {
      
      // first we have n_attr characters 
      // if the i-th is 'n' then that column is NULL
      // as reading strings is quite expensive we use byte arrays...
      
      in.readFully(flags);
      packetlen -= nfields;
   
      
      for (int i=0; i<nfields; ++i) {
	
	

	if (flags[i] == 0x6e) {  // 'n'	  
	  // entry is null
	  data[i] = null;
	  continue;
	}

	  
	// length of entry
	int len = in.readInt() - 4;
	packetlen -= 4;
	
	
	if (len > 0) {
	  // allocate byte array large enough to hold value
	  byte[] value = new byte[len];
	  
	  // read value
	  in.readFully(value);	    
	  packetlen -= len;
	  
	  if (binary)
	    data[i] = value;	    
	  else
	    data[i] = new String(value);
	  
	} else
	  {
	    if (binary)
	      data[i] = null;	  
	    else
	      data[i] = "";
	  }

      } // for

      // skip the remaining System Attributes
      parent.skipBytesFully(packetlen);
      
    } catch ( Exception e ) {
      throw new SQLException("Network error reading results"+e);
    }


    return data;
  }
  
  
  



  /**
   * Helper function to read a result set from the network
   *
   */
  private void readResult(DataInputStream in, int packetlen) throws SQLException
  {

    columns = null;
    tuples = null;

    currentRow = 0;

    // temp:
    int rowcnt=0;

    
    try {

      // First, pull in the attribute table
      // Read number of attributes/fields (4 byte integer)

      int nfields = in.readInt();
      packetlen -= 4;

      // allocate our column datastructure
      columns = new Vector(nfields);

      // loop over fields and collect interesting info

      int currentfield = nfields;
      while ( currentfield-- > 0 ) {

	MIColumn col = new MIColumn();

	// read name
	int namelen = in.readInt() - 4;
	col.name = parent.readstring(namelen);

	// read 64 bit oid
        long oid = in.readLong();

	// map to 32 bit integer which we use as type id.
	col.id = TypeTable.findIndex(oid);

	// bound, parameter, nullable
	// for decimal/numeric bound is precision and parameter is scale.
	col.bound = in.readInt();
	col.parameter = in.readInt();
	col.nullable = in.readInt();


	// System.out.println("Col:"+col.name+" bound:"+bound+" param:"+parameter);

	// read: 4 + namelen + 8 + 3*4  bytes
	packetlen -= 24+namelen;

	// size is not passed on
	col.size = 0;

	columns.addElement(col);
	
	//if (MId.debug)
	//  System.out.println(col);

      }


      if (packetlen != 0) 
	throw new SQLException("packet not exhausted");
      

      
      //
      // now read in the data, they come in ('Q','T') packets,
      // one for each record.
      // 
      // a ('Q','C') packet signals the end.
   


      tuples = new Vector(1000);
      flags = new byte[nfields];


      // Now, pull in the tuple results
      boolean go_on = true;
      while ( go_on ) {

	// read header
	//     word32  length
	//     word16  channel (not used)
	//     char8   class
	//     char8   subclass
	
	packetlen = in.readInt(); 
	short channel = in.readShort(); 
	char classid = (char) in.readByte(); 
	char subclassid = (char) in.readByte(); 
	
	

	if (MId.debug)
	  System.out.println("readResult. Header=('"+classid+"','"+subclassid+"') packetlen="+packetlen+
			     " available: "+in.available());
	
	if (classid != 'Q') 
	  throw new SQLException("('Q','x') packet expected, received ('"+classid+"','"+subclassid+"')");

	
	
	
	switch ( subclassid ) {


	case 'T': // ('Q','T') A tuple

	  // we have another record waiting for us.
	  // first we get the number of attributes 

	  if (MId.debug)
	    System.out.println("Row "+(++rowcnt));

	  int n_attr = in.readInt();
	  packetlen -= 4;
	  if (MId.debug) 
	    System.out.println("Number of attributes: "+n_attr);

	  // we already knew this from the Tuple Descriptor, so check if they match
	  if (n_attr != nfields)
	    throw new SQLException("Column counts are inconsistent");	  

	  // leave rest to readTuple....
	  tuples.addElement(readTuple(in,n_attr,parent.BINARY,packetlen));
	  packetlen = 0;
	  break;


	  
	case 'E':   // ('Q','E') error, string follows.
	  
	  String errmess;
	  if (packetlen > 0) {
	    // try to read errorstring
	    int len = in.readInt();
	    packetlen -=4;
	    errmess = parent.readstring(len);
	    packetlen -= len;
	    if (packetlen>0)
	      parent.skipBytesFully(packetlen);
	  } 
	  else errmess = "?";
	  throw new SQLException("Error: "+errmess);	       

	case 'C':  // ('Q','C') done with query 

	  String mess;
	  if (packetlen > 0) {
	    int len = in.readInt();
	    packetlen -=4;
	    mess = parent.readstring(len);
	    packetlen -= len;
	  } 
	  else mess = "?";

	  if (MId.debug) 
	    System.out.println("Done with query:"+mess);
	  
	  int rows_affected = in.readInt();
	  packetlen -= 4;
	  if (MId.debug) 
	    System.out.println("Rows affected:"+rows_affected);
	  

	  // update count does not need to be set: should be -1.
	  

	  // we are really done
	  go_on = false;
	  break;


	  
	default:
	  throw new SQLException("Network protocol error reading results: "+subclassid);
	  

	} // switch

	if (packetlen>0) 
	  parent.skipBytesFully(packetlen);
	 
      }
    } catch ( Exception e ) {
      throw new SQLException("Network error reading results:"+e);
    }
  }
  
  /**
   * Constructor for a result set
   *
   */
  MIResultSet(MIStatement st, DataInputStream in, int packetlen) throws SQLException
  {
    
    parent = st;

    TypeTable = st.parent.typetable;
    
    //long time = System.currentTimeMillis();
    readResult(in, packetlen);
    //time = System.currentTimeMillis() - time;
 
    //System.out.println("readResult: "+time+" ms.");
    //System.out.println("readTuple: "+tupletime+" ms.");

    meta = new MIResultSetMetaData(this);

  }
  

  /**
   * in case we don't have a result set, eg after update.
   */
  MIResultSet() {
    // Reset the class
    columns = new Vector();
    tuples = new Vector();
    currentRow = 0;
  }


  /**
   * Helper functions to get the current object
   *
   */


  // Object is either a string or a byte array
  private Object value(String columnName) throws SQLException
  {
    int col = findColumn(columnName);
    if (parent.BINARY)
      return TypeTable.ObjectConversion(value(col),
					(MIColumn)columns.elementAt(col-1));
    else
      return value(col);
  }
  
  private Object value(int row,int column) throws SQLException
  {
    try {
      if (parent.BINARY)
	return TypeTable.ObjectConversion(((Object[]) tuples.elementAt(row-1))[column-1],
					  (MIColumn)columns.elementAt(column-1));
      else
	return ((Object[]) tuples.elementAt(row-1))[column-1];
    } catch ( Exception e ) {
      throw new SQLException("Out of result set bounds? value(int, int)" + e );
    }
  }
  
  private Object value(int column) throws SQLException
  {
    try {
      if (parent.BINARY) 
	return TypeTable.ObjectConversion(((Object[])tuples.elementAt(currentRow-1))[column-1],
					  (MIColumn)columns.elementAt(column-1));    
      else
	return ((Object[]) tuples.elementAt(currentRow-1))[column-1];
    } catch ( Exception e ) {
      throw new SQLException("value(int) CR:"+(currentRow-1)+" CL:"+(column-1)+"->" + e );
    }
  }
  
  
  // The isNull() method returns true if the given column in the
  // current row holds the SQL "null".
  
  private boolean isNull(int column) throws SQLException
  {
    return (value(column) == null);
  }
  
  private boolean isNull(String columnName) throws SQLException
  {
    return (value(columnName) == null);
  }
  
  
  // New API (JPM)
  public boolean wasNull() throws SQLException
  {
    return wasException;
  }
  

  // Methods for accessing columns in the current row.
  public String getString(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return null;
      else {	
	if (parent.BINARY)
	  return o.toString();
	else
	  return (String) o;
      }
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  // Methods for accessing columns in the column name.
  public String getString(String columnName) throws SQLException
  {
    try {
      return getString(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }
  }
  
  public boolean getBoolean(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return(false);
      return ((Boolean) o).booleanValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public boolean getBoolean(String columnName) throws SQLException
  {
    try {
      return getBoolean(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }
  }
  
  public byte getByte(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return(0);
      return ((Byte) o).byteValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public byte getByte(String columnName) throws SQLException
  {
    try {
      return getByte(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }

  }
  
  public short getShort(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return 0;
      return ((Short) o).shortValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public short getShort(String columnName) throws SQLException
  {
    try {
      return getShort(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }

  }
  
  public int getInt(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return 0;
      return ((Integer) o).intValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public int getInt(String columnName) throws SQLException
  {
    try {
      return getInt(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }
  }
  
  public long getLong(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return 0;
      return ((Long) o).longValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("getLong(int) conversion error: "+e);
    }
  }
  
  public long getLong(String columnName) throws SQLException
  {
    try {
      return getLong(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }

  }
  
  public float getFloat(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return 0;
      return ((Float) o).floatValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public float getFloat(String columnName) throws SQLException
  {
    try {
      return getFloat(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }

  }
  
  public double getDouble(int column) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return 0;
      return ((Double) o).doubleValue();
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public double getDouble(String columnName) throws SQLException
  {
    try {
      return getDouble(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }

  }
  
  public BigDecimal getBigDecimal(int column, int scale) throws SQLException
  {
    wasException = false;
    try {
      Object o = value(column);
      if (o == null)
	return null;
      return ((BigDecimal) o).setScale(scale);
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public BigDecimal getBigDecimal(String columnName, int scale) throws SQLException
  {
    try {
      return getBigDecimal(findColumn(columnName),scale);
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }
  }
  
  public byte[] getBytes(int column) throws SQLException
  {
    return (byte[]) ((Object[]) tuples.elementAt(currentRow-1))[column-1];
  }
  
  public byte[] getBytes(String columnName) throws SQLException
  {
    return getBytes(findColumn(columnName));
  }
  
  public byte[] getLongBinaryStream(int column) throws SQLException
  {
    return getBytes(column);
  }
  
  public byte[] getLongBinaryStream(String columnName) throws SQLException
  {
    return getBytes(columnName);
  }
  
  public java.sql.Date getDate(int column) throws SQLException
  {
    wasException = false;
    try {
      return (java.sql.Date) value(column);
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }
  }
  
  public java.sql.Date getDate(String columnName) throws SQLException
  {
    try {
      return getDate(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }
  }
  
  public java.sql.Time getTime(int column) throws SQLException
  {
    wasException = false;
    try {
      return (java.sql.Time) value(column);
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }

  }
  
  public java.sql.Time getTime(String columnName) throws SQLException
  {
    try {
      return getTime(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }
  }
  
  public java.sql.Timestamp getTimestamp(int column) throws SQLException
  {
    wasException = false;
    try {
      return (java.sql.Timestamp) value(column);
    } catch ( Exception e ) {
      wasException = true;
      throw new SQLException("Result conversion error: "+e);
    }

  }
  
  public java.sql.Timestamp getTimestamp(String columnName) throws SQLException
  {
    try {
      return getTimestamp(findColumn(columnName));
    } catch(Exception e) {
      throw new SQLException(e.toString());
    }

  }
  
  // The normal getChars and getBinary methods are suitable for
  // reading normal sized data.  However occasionally it may be
  // necessary to access LONGVARCHAR or LONGVARBINARY fields that
  // are multiple Megabytes in size.  To support this case we provide
  // getCharStream and getBinaryStream methods that return Java
  // IO streams from which data can be read in chunks.
  
  public java.io.InputStream getAsciiStream(int column) throws SQLException
  {
    byte[] b = getBytes(column);
    if (b == null)
      return null;
    return new ByteArrayInputStream(b);
  }
  
  public java.io.InputStream getAsciiStream(String columnName) throws SQLException
  {
    return getAsciiStream(findColumn(columnName));
  }
  
  public java.io.InputStream getUnicodeStream(int column) throws SQLException
  {
    return getAsciiStream(column);	// Bad Implementation!
  }
  
  public java.io.InputStream getUnicodeStream(String columnName) throws SQLException
  {
    return getAsciiStream(columnName);
  }
  
  public java.io.InputStream getBinaryStream(int column) throws SQLException
  {
    return getAsciiStream(column);
  }
  
  public java.io.InputStream getBinaryStream(String columnName) throws SQLException
  {
    return getBinaryStream(findColumn(columnName));
  }
  
  // next moves us to the next row of the results.
  // It returns true if if has moved to a valid row, false if
  // all the rows have been processed.
  public boolean next() throws SQLException {
    // MId.h("NEXT!");
    if (tuples == null) {
     MId.h("tuples == null!");
     return false;
    } else
      return ((++currentRow) <= tuples.size());
  }
  
  // Return the number of the current row.  The first row containing
  // actual data is row 1.
  public int getRowNumber() throws SQLException
  {
    return currentRow;
  }
  
  // Return the number of columns in the ResultSet
  public int getColumnCount() throws SQLException
  {
    return columns.size();
  }
  
  // The total number of rows returned by the query.  Note that on some
  // databases this method may be very expensive.
  public int getRowCount() throws SQLException
  {
    return tuples.size();
  }
  
  // close frees up internal state associated with the ResultSet
  // You should normally call close when you are done with a ResultSet
  // ResultSets are also closed automatically when their Statement is closed.
  public void close() throws SQLException
  {
    columns = null;
    tuples = null;
  }
  
  //----------------------------------------------------------------------
  // Advanced features:
  
  // getCursorname returns a cursor name that can be used to identify the
  // current position in the Resultset to a separate statement that is
  // executing a positioned update or positioned delete.  If the database
  // doesn't support positioned delete or update, it may return "" here.
  public String getCursorName() throws SQLException
  {
    //printTuples();
    return (new Integer(currentRow)).toString();
  }
  
  // You can obtain a ResultMetaData object to obtain information
  // about the number, types and properties of the result columns.
  public ResultSetMetaData getMetaData() throws SQLException
  {
    return meta;
  }
  
  public Object getObject(String columnName) throws SQLException
  {
    return(getObject(findColumn(columnName)));
  }
  
  // You can obtain a column result as a Java Object.
  // See the JDBC spec's "Dynamic Programming" chapter for details.
  public Object getObject(int column) throws SQLException {
    return value(column);   
  }
  
  
  public SQLWarning getWarnings() throws SQLException
  {
    return(null);
  }
  
  public void clearWarnings() throws SQLException
  {
    
  }
  
  public int findColumn(String columnName) throws SQLException
  {
    for ( int loop=0; loop < columns.size(); loop++ ) {
      MIColumn col = (MIColumn) columns.elementAt(loop);
      if ( col.name.equals(columnName) ) {
	return (loop + 1);
      }
    }
    throw new SQLException("Couldn't find "+columnName+" in ResultSet");
  }
}
