// Java Interface to Illustra

// Copyright (c) 1997. Erwin Kalvelagen, TNO FEL, <kalvelagen@fel.tno.nl>


// In this class we maintain a type table.
// Basically we store the following:
//     type-id (index)
//     Illustra-oid (long)
//     type-name (string)
//     type-size (int)
//     SQL Type (int)
 
package mijdbc;

import java.util.*;
import java.sql.*;
import java.io.*;
import java.math.*;

/**
 * JDBC Interface to Illustra functions
 */

public class MITypeTable 
{
  
  Vector table;
  static Hashtable TypeDict; // conversion Illustra type -> SQL type

  static {
    TypeDict = new Hashtable();
    // TypeDict.put("abstime",);
    TypeDict.put("boolean", new Integer(Types.BIT));
    TypeDict.put("bytea", new Integer(Types.VARBINARY));
    TypeDict.put("character", new Integer(Types.CHAR));
    TypeDict.put("char", new Integer(Types.CHAR));
    TypeDict.put("char1", new Integer(Types.TINYINT));
    TypeDict.put("date", new Integer(Types.DATE));
    TypeDict.put("decimal", new Integer(Types.DECIMAL));
    TypeDict.put("double precision", new Integer(Types.DOUBLE));
    TypeDict.put("int1", new Integer(Types.TINYINT));
    TypeDict.put("integer", new Integer(Types.INTEGER));
    TypeDict.put("large_text", new Integer(Types.LONGVARCHAR));
    TypeDict.put("numeric", new Integer(Types.NUMERIC));
    TypeDict.put("real", new Integer(Types.REAL));
    //TypeDict.put("reltime",);
    TypeDict.put("smallint", new Integer(Types.SMALLINT));
    TypeDict.put("text", new Integer(Types.VARCHAR));
    TypeDict.put("time", new Integer(Types.TIME));
    TypeDict.put("time_tz", new Integer(Types.TIME));
    TypeDict.put("timestamp", new Integer(Types.TIMESTAMP));
    TypeDict.put("timestamp_tz", new Integer(Types.TIMESTAMP));
    TypeDict.put("varchar", new Integer(Types.VARCHAR));
    
  }
  


  /** 
   * constructor for MITypeTable
   */
  public MITypeTable() {
    table = new Vector();
  }


  /**
   * add a new type to the table. No check is performed if it
   * is already there.
   */
  public void addType(long oid, String typename, int typesize) {
    if (MId.debug)
      System.out.println("Storing type "+typename);
    Object[] record = new Object[4];
    record[0] = new Long(oid);
    record[1] = typename;
    record[2] = new Integer(typesize);
    record[3] = new Integer(SQLType(typename));
    table.addElement(record);
  }


  /**
   * stupid linear search for now
   */
  public int findIndex(long oid) {
    for (int i=0; i<table.size(); ++i) {
      if (oid == getOID(i))
	return(i);
    }
    return -1;
  }

     
  public String getTypeName(int index) {
    Object[] record = (Object[])table.elementAt(index);
    return (String)record[1];
  }

  public String getTypeName(long oid) {
    int index = findIndex(oid);
    if (index < 0) 
      return "";
    else
      return getTypeName(index);
  }

  public int getSQLType(int index) {
    Object[] record = (Object[])table.elementAt(index);
    return ((Integer)record[3]).intValue();
  }

  public long getOID(int index) {
    Object[] record = (Object[])table.elementAt(index);
    return ((Long)record[0]).longValue();
  }

  public int SQLType(String IllustraType) {
    Object o = TypeDict.get(IllustraType);
    if (o == null)
      return(Types.OTHER);
    return ((Integer)o).intValue();
  }


  //----------------------------------------------------------------
  // This is an important beast.
  // Coming from the database, all objects are stored as simple
  // byte arrays. This method looks at the type of the relevant 
  // column, and returns the correct type, casted into an Object.
  //---------------------------------------------------------------- 
  public Object ObjectConversion(Object ovalue,  MIColumn col) throws SQLException {


    try {
      
      if (ovalue == null)
	return null;
      
      int index = col.id;
      
      
      //System.out.println("  SQLType ="+getSQLType(index)+"  "+getTypeName(index));
      
      byte[] value = (byte[]) ovalue;
      
      
      //MIHexDump.DumpBuffer(value);
      
      switch(getSQLType(index)) {
	
	
      case Types.BIT:
	{
	  //System.out.println("  SQLType = BIT");
	  if (value.length != 1)
	    throw new SQLException("boolean expects 1 byte length");
	  switch (value[0]) {
	  case 0: 
	    return new Boolean(false);
	  case 1:
	    return new Boolean(true);
	  default:
	    throw new SQLException("unexpected boolean value");
	  }
	}

      case Types.VARBINARY:
	{
	  int len =  ((value[0] & 0xff) << 24) + 
	    ((value[1] & 0xff) << 16) + 
	    ((value[2] & 0xff) << 8) + 
	    (value[3] & 0xff) - 4;
	  byte b[] = new byte[len];
	  for (int i=0; i<len; ++i)
	    b[i] = value[i+4];
	  return b;
	}


      case Types.LONGVARCHAR:
      case Types.VARCHAR:
      case Types.CHAR:
	{
	  // first an integer containing containing number of chars + 4
	  // then the chars
	  int len =  ((value[0] & 0xff) << 24) + 
	    ((value[1] & 0xff) << 16) + 
	    ((value[2] & 0xff) << 8) + 
	    (value[3] & 0xff);
	  StringBuffer s = new StringBuffer();
	  for (int i=4; i < len; ++i)
	    s.append((char)value[i]);
	  return new String(s);
	}
      
      case Types.TINYINT:
	// either a char1 or an int1. Interpret char1 as a character
	// and int1 as a byte.
	if (getTypeName(index).equals("char1")) 
	  return new Character((char)value[0]);
	else { // int1
	  //System.out.print("(int1->Byte)");
	  return new Byte(value[0]);
	}
	
       
      case Types.DATE:
	int year = ((value[0] & 0xff) << 8) + (value[1] & 0xff) - 1900;
	int month = value[2]-1;
	int day = value[3];
	return  new java.sql.Date(year, month, day);
	
      case Types.DECIMAL:
      case Types.NUMERIC: 
	{
	  // first an integer containing containing number of chars + 4
	  // then the bytes
	  int len =  ((value[0] & 0xff) << 24) + 
	    ((value[1] & 0xff) << 16) + 
	    ((value[2] & 0xff) << 8) + 
	    (value[3] & 0xff) - 4;
	  byte[] data = new byte[len];
	  for (int k=0; k<len; ++k)
	    data[k] = value[k+4];
	  BigInteger bigint = new BigInteger(data);
	  int scale = col.parameter;
	  return new BigDecimal(bigint, scale);
	}


      case Types.DOUBLE:
	{
	  DataInputStream is = new DataInputStream( new ByteArrayInputStream(value));
	  return new Double(is.readDouble());
	}
	

      case Types.INTEGER:
	return new Integer( 
			   ((value[0] & 0xff) << 24) + 
			   ((value[1] & 0xff) << 16) + 
			   ((value[2] & 0xff) << 8) + 
			   (value[3] & 0xff)
			   );
	
	  

      case Types.REAL:
	return new Float(
			 Float.intBitsToFloat( 
					      ((value[0] & 0xff) << 24) + 
					      ((value[1] & 0xff) << 16) + 
					      ((value[2] & 0xff) << 8) + 
					      (value[3] & 0xff)
					      )
			 );
	
	  



      case Types.SMALLINT:
	return new Short((short)(((value[0] & 0xff) << 8) + (value[1] & 0xff)));
	
      case Types.TIME:
	int hour = value[0] & 0xff;
	int minute = value[1] & 0xff;
	int second = value[2] & 0xff;
	int tz_hour = value[3] & 0xff;
	int tz_min = value[4] & 0xff;
	// int tz_fraction = ((value[5] & 0xff) << 16) + ((value[6] & 0xff) << 8) +  (value[7] & 0xff);
	// System.out.println("h,m,s:"+hour+":"+minute+":"+second+" tz_hour,tz_min:"+tz_hour+":"+tz_min);
	// sloppy: forget about tz_fraction
	minute += tz_min;
	if (minute > 59) {
	  minute -= 60;
	  ++hour;
	}
	if (minute < 0) {
	  minute += 60;
	  --hour;
	}
	hour += tz_hour;
	if (hour > 23)
	  hour -= 24;
	if (hour < 0)
	  hour += 24;
	return new java.sql.Time(hour,minute,second);
      
      case Types.TIMESTAMP:	
	year = ((value[0] & 0xff) << 8) + (value[1] & 0xff) - 1900;
	month = value[2]-1;
	day = value[3];
	hour = value[4] & 0xff;
	minute = value[5] & 0xff;
	second = value[6] & 0xff;
	tz_hour = value[7] & 0xff;
	tz_min = value[8] & 0xff;
	// int tz_fraction = ((value[9] & 0xff) << 16) + ((value[10] & 0xff) << 8) +  (value[11] & 0xff);
	// sloppy: forget about tz_fraction
	java.sql.Timestamp ts = new java.sql.Timestamp(year,month,day,hour,minute,second,0);
	long t = ts.getTime();
	t += 1000*60*(tz_hour*60+tz_min);
	ts.setTime(t);
	return ts;

      default:
	return ovalue;


      }  // switch
      
    } catch (Exception e) {
      throw new SQLException(""+e);
    }
    
  }

}

